/*
The Pattern has 2 kind of wildcards:
  *     wildcard for any (including zero) number of characters
  ?     wildcard for exactly one single character
Handle '*' :
  skip 0 .. remaining_length characters, and check rest of match
Handle '?' :
  skip next character (if available)
Others:
  check 1 character
*/

class PatternMatch
{
  public void                   PatternMatch()
  {
  }

  static public boolean         Match(
  String                        oString,
  String                        oPattern )
  {
    int                         LengthString  = oString.length();
    int                         LengthPattern = oPattern.length();

    if ( LengthPattern == 0 )
    {
      return LengthString == 0;
    }

    int                         iStar   = oPattern.indexOf( "*" );
    int                         iSingle = oPattern.indexOf( "?" );
    int                         iFirst;
    String                      lPattern1 = oPattern.substring( 1 );

    if ( iSingle == 0 )
    { // Pattern starts with '?'.
      if ( LengthString == 0 )
      {
        return false;
      }
      return Match( oString.substring( 1 ), lPattern1 );
    }

    if ( iStar == 0 )
    {
      for ( int i = 0; i <= LengthString; i++ )
      {
        if ( Match( oString.substring( i ), lPattern1 ) )
        {
          return true;
        }
      }
      return false;
    }

    iFirst = LengthPattern;
    if ( iSingle >= 0 )
    {
      iFirst = iSingle;
    }
    if ( ( iStar >= 0 ) && ( iStar < iFirst ) )
    {
      iFirst = iStar;
    }

    if ( ! oString.startsWith( oPattern.substring( 0, iFirst ) ) )
    {
      return false;
    }

    return Match( oString.substring( iFirst ), oPattern.substring( iFirst ) );
  }

  static public String []       MatchList(
  String []                     oList,
  String                        oPattern )
  {
    int                         Length = oList.length;
    boolean []                  baMatch = new boolean[ Length ];
    int                         NewLength = 0;

    for ( int i = 0; i < Length; i++ )
    {
      baMatch[ i ] = Match( oList[i], oPattern );
      if ( baMatch[ i ] )
      {
        NewLength++;
      }
    }

    String []                   laNewList = new String[ NewLength ];
    int                         j = 0;

    for ( int i = 0; i < Length; i++ )
    {
      if ( baMatch[ i ] )
      {
        laNewList[ j++ ] = oList[i];
      }
    }

    return laNewList;
  }

  static public void            Test(
  String                        oString,
  String                        oPattern )
  {
    boolean                     bMatch = Match( oString, oPattern );
    
    System.out.println( "I: " + oString + "  P: " + oPattern + "  " +
                        ((bMatch) ? "YES" : "NO") );
  }

  static public void            TestList(
  String                        oPattern )
  {
    String []                   lList =
    { "a",
      "abc",
      "aaa",
      "abcd",
      "abcc",
      "ac" };

    String []                   oMatches = MatchList( lList, oPattern );

    System.out.println( "P:  " + oPattern );
    for ( int k = 0; k < oMatches.length; k++ )
    {
      System.out.println( "[" + k + "] = " + oMatches[ k ] );
    }
  }

  public static void            main(
  String []                     args )
  {
    Test( "", "*" );
    Test( "", "?" );
    Test( "abc", "a*c" );
    Test( "abc", "a?c" );
    Test( "abc", "a?c*" );
    Test( "abc", "a?c*" );
    Test( "abc", "a??" );
    Test( "abc", "d*c" );
    Test( "abc", "**c" );
    Test( "abc", "?*c" );

    TestList( "a*c" );
    TestList( "a?c" );
  }
}

